"""Modul which implements class and functions
all about scaling images.
"""

import abc
import enum

import ueberzug.geometry as geometry


class ImageScaler(metaclass=abc.ABCMeta):
    """Describes the structure used to define image scaler classes.

    Defines a general interface used to implement different ways
    of scaling images to specific sizes.
    """

    @staticmethod
    @abc.abstractmethod
    def get_scaler_name():
        """Returns:
        str: the constant name which is associated to this scaler.
        """
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def is_indulgent_resizing():
        """This method specifies whether the
        algorithm returns noticeable different results for
        the same image with different sizes (bigger than the
        maximum size which is passed to the scale method).

        Returns:
            bool: False if the results differ
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def calculate_resolution(self, image, width: int, height: int):
        """Calculates the final resolution of the scaled image.

        Args:
            image (PIL.Image): the image which should be scaled
            width (int): maximum width that can be taken
            height (int): maximum height that can be taken

        Returns:
            tuple: final width: int, final height: int
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def scale(self, image, position: geometry.Point, width: int, height: int):
        """Scales the image according to the respective implementation.

        Args:
            image (PIL.Image): the image which should be scaled
            position (geometry.Position): the centered position, if possible
                Specified as factor of the image size,
                so it should be an element of [0, 1].
            width (int): maximum width that can be taken
            height (int): maximum height that can be taken

        Returns:
            PIL.Image: the scaled image
        """
        raise NotImplementedError()


class OffsetImageScaler(ImageScaler, metaclass=abc.ABCMeta):
    """Extension of the ImageScaler class by Offset specific functions."""

    # pylint can't detect abstract subclasses
    # pylint: disable=abstract-method

    @staticmethod
    def get_offset(position: float, target_size: float, image_size: float):
        """Calculates a offset which contains the position
        in a range from offset to offset + target_size.

        Args:
            position (float): the centered position, if possible
                Specified as factor of the image size,
                so it should be an element of [0, 1].
            target_size (int): the image size of the wanted result
            image_size (int): the image size

        Returns:
            int: the offset
        """
        return int(
            min(
                max(0, position * image_size - target_size / 2),
                image_size - target_size,
            )
        )


class MinSizeImageScaler(ImageScaler):
    """Partial implementation of an ImageScaler.
    Subclasses calculate the final resolution of the scaled image
    as the minimum value of the image size and the maximum size.
    """

    # pylint: disable=abstract-method

    def calculate_resolution(self, image, width: int, height: int):
        return (min(width, image.width), min(height, image.height))


class CropImageScaler(MinSizeImageScaler, OffsetImageScaler):
    """Implementation of the ImageScaler
    which crops out the maximum image size.
    """

    @staticmethod
    def get_scaler_name():
        return "crop"

    @staticmethod
    def is_indulgent_resizing():
        return False

    def scale(self, image, position: geometry.Point, width: int, height: int):
        width, height = self.calculate_resolution(image, width, height)
        image_width, image_height = image.width, image.height
        offset_x = self.get_offset(position.x, width, image_width)
        offset_y = self.get_offset(position.y, height, image_height)
        return image.crop(
            (offset_x, offset_y, offset_x + width, offset_y + height)
        )


class DistortImageScaler(ImageScaler):
    """Implementation of the ImageScaler
    which distorts the image to the maximum image size.
    """

    @staticmethod
    def get_scaler_name():
        return "distort"

    @staticmethod
    def is_indulgent_resizing():
        return True

    def calculate_resolution(self, image, width: int, height: int):
        return width, height

    def scale(self, image, position: geometry.Point, width: int, height: int):
        import PIL.Image

        width, height = self.calculate_resolution(image, width, height)
        return image.resize((width, height), PIL.Image.LANCZOS)


class FitContainImageScaler(DistortImageScaler):
    """Implementation of the ImageScaler
    which resizes the image that either
    the width matches the maximum width
    or the height matches the maximum height
    while keeping the image ratio.
    """

    @staticmethod
    def get_scaler_name():
        return "fit_contain"

    @staticmethod
    def is_indulgent_resizing():
        return True

    def calculate_resolution(self, image, width: int, height: int):
        factor = min(width / image.width, height / image.height)
        return int(image.width * factor), int(image.height * factor)


class ContainImageScaler(FitContainImageScaler):
    """Implementation of the ImageScaler
    which resizes the image to a size <= the maximum size
    while keeping the image ratio.
    """

    @staticmethod
    def get_scaler_name():
        return "contain"

    @staticmethod
    def is_indulgent_resizing():
        return True

    def calculate_resolution(self, image, width: int, height: int):
        return super().calculate_resolution(
            image, min(width, image.width), min(height, image.height)
        )


class ForcedCoverImageScaler(DistortImageScaler, OffsetImageScaler):
    """Implementation of the ImageScaler
    which resizes the image to cover the entire area which should be filled
    while keeping the image ratio.
    If the image is smaller than the desired size
    it will be stretched to reach the desired size.
    If the ratio of the area differs
    from the image ratio the edges will be cut off.
    """

    @staticmethod
    def get_scaler_name():
        return "forced_cover"

    @staticmethod
    def is_indulgent_resizing():
        return True

    def scale(self, image, position: geometry.Point, width: int, height: int):
        import PIL.Image

        width, height = self.calculate_resolution(image, width, height)
        image_width, image_height = image.width, image.height
        if width / image_width > height / image_height:
            image_height = int(image_height * width / image_width)
            image_width = width
        else:
            image_width = int(image_width * height / image_height)
            image_height = height
        offset_x = self.get_offset(position.x, width, image_width)
        offset_y = self.get_offset(position.y, height, image_height)

        return image.resize(
            (image_width, image_height), PIL.Image.LANCZOS
        ).crop((offset_x, offset_y, offset_x + width, offset_y + height))


class CoverImageScaler(MinSizeImageScaler, ForcedCoverImageScaler):
    """The same as ForcedCoverImageScaler but images won't be stretched
    if they are smaller than the area which should be filled.
    """

    @staticmethod
    def get_scaler_name():
        return "cover"

    @staticmethod
    def is_indulgent_resizing():
        return True


@enum.unique
class ScalerOption(str, enum.Enum):
    """Enum which lists the useable ImageScaler classes."""

    DISTORT = DistortImageScaler
    CROP = CropImageScaler
    FIT_CONTAIN = FitContainImageScaler
    CONTAIN = ContainImageScaler
    FORCED_COVER = ForcedCoverImageScaler
    COVER = CoverImageScaler

    def __new__(cls, scaler_class):
        inst = str.__new__(cls)
        # Based on an official example
        # https://docs.python.org/3/library/enum.html#using-a-custom-new
        # So.. stfu pylint
        # pylint: disable=protected-access
        inst._value_ = scaler_class.get_scaler_name()
        inst.scaler_class = scaler_class
        return inst



import abc
import enum
import os.path

import attr

import ueberzug.geometry as geometry
import ueberzug.scaling as scaling
import ueberzug.conversion as conversion


@attr.s
class Action(metaclass=abc.ABCMeta):
    """Describes the structure used to define actions classes.

    Defines a general interface used to implement the building of commands
    and their execution.
    """

    action = attr.ib(
        type=str,
        default=attr.Factory(
            lambda self: self.get_action_name(), takes_self=True
        ),
    )

    @staticmethod
    @abc.abstractmethod
    def get_action_name():
        """Returns the constant name which is associated to this action."""
        raise NotImplementedError()

    @abc.abstractmethod
    async def apply(self, windows, view, tools):
        """Executes the action on  the passed view and windows."""
        raise NotImplementedError()


@attr.s(kw_only=True)
class Drawable:
    """Defines the attributes of drawable actions."""

    draw = attr.ib(default=True, converter=conversion.to_bool)
    synchronously_draw = attr.ib(default=False, converter=conversion.to_bool)


@attr.s(kw_only=True)
class Identifiable:
    """Defines the attributes of actions
    which are associated to an identifier.
    """

    identifier = attr.ib(type=str)


@attr.s(kw_only=True)
class DrawAction(Action, Drawable, metaclass=abc.ABCMeta):
    """Defines actions which redraws all windows."""

    # pylint: disable=abstract-method
    __redraw_scheduled = False

    @staticmethod
    def schedule_redraw(windows):
        """Creates a async function which redraws every window
        if there is no unexecuted function
        (returned by this function)
        which does the same.

        Args:
            windows (batch.BatchList of ui.CanvasWindow):
                the windows to be redrawn

        Returns:
            function: the redraw function or None
        """
        if not DrawAction.__redraw_scheduled:
            DrawAction.__redraw_scheduled = True

            async def redraw():
                windows.draw()
                DrawAction.__redraw_scheduled = False

            return redraw()
        return None

    async def apply(self, windows, view, tools):
        if self.draw:
            import asyncio

            if self.synchronously_draw:
                windows.draw()
                # force coroutine switch
                await asyncio.sleep(0)
                return

            function = self.schedule_redraw(windows)
            if function:
                asyncio.ensure_future(function)


@attr.s(kw_only=True)
class ImageAction(DrawAction, Identifiable, metaclass=abc.ABCMeta):
    """Defines actions which are related to images."""

    # pylint: disable=abstract-method
    pass


@attr.s(kw_only=True)
class AddImageAction(ImageAction):
    """Displays the image according to the passed option.
    If there's already an image with the given identifier
    it's going to be replaced.
    """

    x = attr.ib(type=int, converter=int)
    y = attr.ib(type=int, converter=int)
    path = attr.ib(type=str)
    width = attr.ib(type=int, converter=int, default=0)
    height = attr.ib(type=int, converter=int, default=0)
    scaling_position_x = attr.ib(type=float, converter=float, default=0)
    scaling_position_y = attr.ib(type=float, converter=float, default=0)
    scaler = attr.ib(
        type=str, default=scaling.ContainImageScaler.get_scaler_name()
    )
    # deprecated
    max_width = attr.ib(type=int, converter=int, default=0)
    max_height = attr.ib(type=int, converter=int, default=0)

    @staticmethod
    def get_action_name():
        return "add"

    def __attrs_post_init__(self):
        self.width = self.max_width or self.width
        self.height = self.max_height or self.height
        # attrs doesn't support overriding the init method
        # pylint: disable=attribute-defined-outside-init
        self.__scaler_class = None
        self.__last_modified = None

    @property
    def scaler_class(self):
        """scaling.ImageScaler: the used scaler class of this placement"""
        if self.__scaler_class is None:
            self.__scaler_class = scaling.ScalerOption(self.scaler).scaler_class
        return self.__scaler_class

    @property
    def last_modified(self):
        """float: the last modified time of the image"""
        if self.__last_modified is None:
            self.__last_modified = os.path.getmtime(self.path)
        return self.__last_modified

    def is_same_image(self, old_placement):
        """Determines whether the placement contains the same image
        after applying the changes of this command.

        Args:
            old_placement (ui.CanvasWindow.Placement):
                the old data of the placement

        Returns:
            bool: True if it's the same file
        """
        return old_placement and not (
            old_placement.last_modified < self.last_modified
            or self.path != old_placement.path
        )

    def is_full_reload_required(
        self, old_placement, screen_columns, screen_rows
    ):
        """Determines whether it's required to fully reload
        the image of the placement to properly render the placement.

        Args:
            old_placement (ui.CanvasWindow.Placement):
                the old data of the placement
            screen_columns (float):
                the maximum amount of columns the screen can display
            screen_rows (float):
                the maximum amount of rows the screen can display

        Returns:
            bool: True if the image should be reloaded
        """
        return old_placement and (
            (
                not self.scaler_class.is_indulgent_resizing()
                and old_placement.scaler.is_indulgent_resizing()
            )
            or (old_placement.width <= screen_columns < self.width)
            or (old_placement.height <= screen_rows < self.height)
        )

    def is_partly_reload_required(
        self, old_placement, screen_columns, screen_rows
    ):
        """Determines whether it's required to partly reload
        the image of the placement to render the placement more quickly.

        Args:
            old_placement (ui.CanvasWindow.Placement):
                the old data of the placement
            screen_columns (float):
                the maximum amount of columns the screen can display
            screen_rows (float):
                the maximum amount of rows the screen can display

        Returns:
            bool: True if the image should be reloaded
        """
        return old_placement and (
            (
                self.scaler_class.is_indulgent_resizing()
                and not old_placement.scaler.is_indulgent_resizing()
            )
            or (self.width <= screen_columns < old_placement.width)
            or (self.height <= screen_rows < old_placement.height)
        )

    async def apply(self, windows, view, tools):
        try:
            import ueberzug.ui as ui
            import ueberzug.loading as loading

            old_placement = view.media.pop(self.identifier, None)
            cache = old_placement and old_placement.cache
            image = old_placement and old_placement.image

            max_font_width = max(
                map(lambda i: i or 0, windows.parent_info.font_width or [0])
            )
            max_font_height = max(
                map(lambda i: i or 0, windows.parent_info.font_height or [0])
            )
            font_size_available = max_font_width and max_font_height
            screen_columns = (
                font_size_available and view.screen_width / max_font_width
            )
            screen_rows = (
                font_size_available and view.screen_height / max_font_height
            )

            # By default images are only stored up to a resolution which
            # is about as big as the screen resolution.
            # (loading.CoverPostLoadImageProcessor)
            # The principle of spatial locality does not apply to
            # resize operations of images with big resolutions
            # which is why those operations should be applied
            # to a resized version of those images.
            # Sometimes we still need all pixels e.g.
            # if the image scaler crop is used.
            # So sometimes it's required to fully load them
            # and sometimes it's not required anymore which is
            # why they should be partly reloaded
            # (to speed up the resize operations again).
            if (
                not self.is_same_image(old_placement)
                or (
                    font_size_available
                    and self.is_full_reload_required(
                        old_placement, screen_columns, screen_rows
                    )
                )
                or (
                    font_size_available
                    and self.is_partly_reload_required(
                        old_placement, screen_columns, screen_rows
                    )
                )
            ):
                upper_bound_size = None
                image_post_load_processor = None
                if (
                    self.scaler_class != scaling.CropImageScaler
                    and font_size_available
                ):
                    upper_bound_size = (
                        max_font_width * self.width,
                        max_font_height * self.height,
                    )
                if (
                    self.scaler_class != scaling.CropImageScaler
                    and font_size_available
                    and self.width <= screen_columns
                    and self.height <= screen_rows
                ):
                    image_post_load_processor = (
                        loading.CoverPostLoadImageProcessor(
                            view.screen_width, view.screen_height
                        )
                    )
                image = tools.loader.load(
                    self.path, upper_bound_size, image_post_load_processor
                )
                cache = None

            view.media[self.identifier] = ui.CanvasWindow.Placement(
                self.x,
                self.y,
                self.width,
                self.height,
                geometry.Point(
                    self.scaling_position_x, self.scaling_position_y
                ),
                self.scaler_class(),
                self.path,
                image,
                self.last_modified,
                cache,
            )
        finally:
            await super().apply(windows, view, tools)


@attr.s(kw_only=True)
class RemoveImageAction(ImageAction):
    """Removes the image with the passed identifier."""

    @staticmethod
    def get_action_name():
        return "remove"

    async def apply(self, windows, view, tools):
        try:
            if self.identifier in view.media:
                del view.media[self.identifier]
        finally:
            await super().apply(windows, view, tools)


@enum.unique
class Command(str, enum.Enum):
    ADD = AddImageAction
    REMOVE = RemoveImageAction

    def __new__(cls, action_class):
        inst = str.__new__(cls)
        inst._value_ = action_class.get_action_name()
        inst.action_class = action_class
        return inst



import atexit
import sys
import os
import asyncio
import signal
import pathlib
import re
import tempfile

import ueberzug.thread as thread
import ueberzug.files as files
import ueberzug.xutil as xutil
import ueberzug.parser as parser
import ueberzug.ui as ui
import ueberzug.batch as batch
import ueberzug.action as action
import ueberzug.tmux_util as tmux_util
import ueberzug.geometry as geometry
import ueberzug.loading as loading
import ueberzug.X as X


async def process_xevents(loop, display, windows):
    """Coroutine which processes X11 events"""
    async for _ in xutil.Events(loop, display):
        if not any(windows.process_event()):
            display.discard_event()


async def process_commands(
    loop, shutdown_routine_factory, windows, view, tools
):
    """Coroutine which processes the input of stdin"""
    try:
        async for line in files.LineReader(loop, sys.stdin):
            if not line:
                break

            try:
                line = re.sub("\\\\", "", line)

                data = tools.parser.parse(line[:-1])
                command = action.Command(data["action"])
                await command.action_class(**data).apply(windows, view, tools)
            except (OSError, KeyError, ValueError, TypeError) as error:
                tools.error_handler(error)
    finally:
        asyncio.ensure_future(shutdown_routine_factory())


async def query_windows(display: X.Display, window_factory, windows, view):
    """Signal handler for SIGUSR1.
    Searches for added and removed tmux clients.
    Added clients: additional windows will be mapped
    Removed clients: existing windows will be destroyed
    """
    parent_window_infos = xutil.get_parent_window_infos(display)
    view.offset = tmux_util.get_offset()
    map_parent_window_id_info = {
        info.window_id: info for info in parent_window_infos
    }
    parent_window_ids = map_parent_window_id_info.keys()
    map_current_windows = {window.parent_id: window for window in windows}
    current_window_ids = map_current_windows.keys()
    diff_window_ids = parent_window_ids ^ current_window_ids
    added_window_ids = diff_window_ids & parent_window_ids
    removed_window_ids = diff_window_ids & current_window_ids
    draw = added_window_ids or removed_window_ids

    if added_window_ids:
        windows += window_factory.create(
            *[map_parent_window_id_info.get(wid) for wid in added_window_ids]
        )

    if removed_window_ids:
        windows -= [map_current_windows.get(wid) for wid in removed_window_ids]

    if draw:
        windows.draw()


async def reset_terminal_info(windows):
    """Signal handler for SIGWINCH.
    Resets the terminal information of all windows.
    """
    windows.reset_terminal_info()


async def shutdown(loop):
    try:
        all_tasks = asyncio.all_tasks()
        current_task = asyncio.current_task()
    except AttributeError:
        all_tasks = asyncio.Task.all_tasks()
        current_task = asyncio.tasks.Task.current_task()

    tasks = [task for task in all_tasks if task is not current_task]
    list(map(lambda task: task.cancel(), tasks))
    await asyncio.gather(*tasks, return_exceptions=True)
    loop.stop()


def shutdown_factory(loop):
    return lambda: asyncio.ensure_future(shutdown(loop))


def setup_tmux_hooks():
    """Registers tmux hooks which are
    required to notice a change in the visibility
    of the pane this program runs in.
    Also it's required to notice new tmux clients
    displaying our pane.

    Returns:
        function which unregisters the registered hooks
    """
    events = (
        "client-session-changed",
        "session-window-changed",
        "pane-mode-changed",
        "client-detached",
    )

    xdg_cache_dir = os.environ.get("XDG_CACHE_HOME")
    if xdg_cache_dir:
        lock_directory_path = pathlib.PosixPath.joinpath(
            pathlib.PosixPath(xdg_cache_dir), pathlib.PosixPath("ueberzug")
        )
    else:
        lock_directory_path = pathlib.PosixPath.joinpath(
            pathlib.PosixPath.home(), pathlib.PosixPath(".cache/ueberzug")
        )

    lock_file_path = lock_directory_path / tmux_util.get_session_id()
    own_pid = str(os.getpid())
    command_template = "ueberzug query_windows "

    try:
        lock_directory_path.mkdir()
    except FileExistsError:
        pass

    def update_hooks(pid_file, pids):
        pids = " ".join(pids)
        command = command_template + pids

        pid_file.seek(0)
        pid_file.truncate()
        pid_file.write(pids)
        pid_file.flush()

        for event in events:
            if pids:
                tmux_util.register_hook(event, command)
            else:
                tmux_util.unregister_hook(event)

    def remove_hooks():
        """Removes the hooks registered by the outer function."""
        with files.lock(lock_file_path) as lock_file:
            pids = set(lock_file.read().split())
            pids.discard(own_pid)
            update_hooks(lock_file, pids)

    with files.lock(lock_file_path) as lock_file:
        pids = set(lock_file.read().split())
        pids.add(own_pid)
        update_hooks(lock_file, pids)

    return remove_hooks


def error_processor_factory(parser):
    def wrapper(exception):
        return process_error(parser, exception)

    return wrapper


def process_error(parser, exception):
    print(
        parser.unparse(
            {
                "type": "error",
                "name": type(exception).__name__,
                "message": str(exception),
                # 'stack': traceback.format_exc()
            }
        ),
        file=sys.stderr,
    )


class View:
    """Data class which holds meta data about the screen"""

    def __init__(self):
        self.offset = geometry.Distance()
        self.media = {}
        self.screen_width = 0
        self.screen_height = 0


class Tools:
    """Data class which holds helper functions, ..."""

    def __init__(self, loader, parser, error_handler):
        self.loader = loader
        self.parser = parser
        self.error_handler = error_handler


def main(options):
    if options["--silent"]:
        try:
            outfile = os.open(os.devnull, os.O_WRONLY)
            os.close(sys.stderr.fileno())
            os.dup2(outfile, sys.stderr.fileno())
        finally:
            os.close(outfile)

    loop = asyncio.get_event_loop()
    executor = thread.DaemonThreadPoolExecutor(max_workers=2)
    parser_object = parser.ParserOption(options["--parser"]).parser_class()
    image_loader = loading.ImageLoaderOption(options["--loader"]).loader_class()
    error_handler = error_processor_factory(parser_object)
    view = View()
    tools = Tools(image_loader, parser_object, error_handler)
    X.init_threads()
    display = X.Display()
    window_factory = ui.CanvasWindow.Factory(display, view)
    window_infos = xutil.get_parent_window_infos(display)
    windows = batch.BatchList(window_factory.create(*window_infos))
    image_loader.register_error_handler(error_handler)
    view.screen_width = display.screen_width
    view.screen_height = display.screen_height

    if tmux_util.is_used():
        atexit.register(setup_tmux_hooks())
        view.offset = tmux_util.get_offset()

    with windows, image_loader:
        loop.set_default_executor(executor)

        for sig in (signal.SIGINT, signal.SIGTERM):
            loop.add_signal_handler(sig, shutdown_factory(loop))

        loop.add_signal_handler(
            signal.SIGUSR1,
            lambda: asyncio.ensure_future(
                query_windows(display, window_factory, windows, view)
            ),
        )

        loop.add_signal_handler(
            signal.SIGWINCH,
            lambda: asyncio.ensure_future(reset_terminal_info(windows)),
        )

        asyncio.ensure_future(process_xevents(loop, display, windows))
        asyncio.ensure_future(
            process_commands(loop, shutdown_factory(loop), windows, view, tools)
        )

        try:
            loop.run_forever()
        finally:
            loop.close()
            executor.shutdown(wait=False)



import abc
import queue
import weakref
import os
import threading
import concurrent.futures
import enum

import ueberzug.thread as thread
import ueberzug.pattern as pattern


INDEX_ALPHA_CHANNEL = 3


def load_image(path, upper_bound_size):
    """Loads the image and converts it
    if it doesn't use the RGB or RGBX mode.

    Args:
        path (str): the path of the image file
        upper_bound_size (tuple of (width: int, height: int)):
            the maximal size to load data for

    Returns:
        tuple of (PIL.Image, bool): rgb image, downscaled

    Raises:
        OSError: for unsupported formats
    """
    import PIL.Image

    image = PIL.Image.open(path)
    original_size = image.width, image.height
    downscaled = False
    mask = None

    if upper_bound_size:
        upper_bound_size = tuple(
            min(size for size in size_pair if size > 0)
            for size_pair in zip(upper_bound_size, original_size)
        )
        image.draft(None, upper_bound_size)
        downscaled = (image.width, image.height) < original_size

    image.load()

    if (
        image.format == "PNG"
        and image.mode in ("L", "P")
        and "transparency" in image.info
    ):
        # Prevent pillow to print the warning
        # 'Palette images with Transparency expressed in bytes should be
        #  converted to RGBA images'
        image = image.convert("RGBA")

    if image.mode == "RGBA":
        mask = image.split()[INDEX_ALPHA_CHANNEL]

    if image.mode not in ("RGB", "RGBX"):
        image_rgb = PIL.Image.new("RGB", image.size, color=(255, 255, 255))
        image_rgb.paste(image, mask=mask)
        image = image_rgb

    return image, downscaled


class ImageHolder:
    """Holds the reference of an image.
    It serves as bridge between image loader and image user.
    """

    def __init__(self, path, image=None):
        self.path = path
        self.image = image
        self.waiter = threading.Condition()

    def reveal_image(self, image):
        """Assigns an image to this holder and
        notifies waiting image users about it.

        Args:
            image (PIL.Image): the loaded image
        """
        with self.waiter:
            self.image = image
            self.waiter.notify_all()

    def await_image(self):
        """Waits till an image loader assigns the image
        if it's not already happened.

        Returns:
            PIL.Image: the image assigned to this holder
        """
        if self.image is None:
            with self.waiter:
                if self.image is None:
                    self.waiter.wait()
        return self.image


class PostLoadImageProcessor(metaclass=abc.ABCMeta):
    """Describes the structure used to define callbacks which
    will be invoked after loading an image.
    """

    @abc.abstractmethod
    def on_loaded(self, image):
        """Postprocessor of an loaded image.
        The returned image will be assigned to the image holder.

        Args:
            image (PIL.Image): the loaded image

        Returns:
            PIL.Image:
                the image which will be assigned
                to the image holder of this loading process
        """
        raise NotImplementedError()


class CoverPostLoadImageProcessor(PostLoadImageProcessor):
    """Implementation of PostLoadImageProcessor
    which resizes an image (if possible -> needs to be bigger)
    such that it covers only just a given resolution.
    """

    def __init__(self, width, height):
        self.width = width
        self.height = height

    def on_loaded(self, image):
        import PIL.Image

        resize_ratio = max(
            min(1, self.width / image.width), min(1, self.height / image.height)
        )

        if resize_ratio != 1:
            image = image.resize(
                (
                    int(resize_ratio * image.width),
                    int(resize_ratio * image.height),
                ),
                PIL.Image.LANCZOS,
            )

        return image


class ImageLoader(metaclass=abc.ABCMeta):
    """Describes the structure used to define image loading strategies.

    Defines a general interface used to implement different ways
    of loading images.
    E.g. loading images asynchron
    """

    @pattern.LazyConstant
    def PLACEHOLDER():
        """PIL.Image: fallback image for occuring errors"""
        # pylint: disable=no-method-argument,invalid-name
        import PIL.Image

        return PIL.Image.new("RGB", (1, 1))

    @staticmethod
    @abc.abstractmethod
    def get_loader_name():
        """Returns the constant name which is associated to this loader."""
        raise NotImplementedError()

    def __init__(self):
        self.error_handler = None

    @abc.abstractmethod
    def load(self, path, upper_bound_size, post_load_processor=None):
        """Starts the image loading procedure for the passed path.
        How and when an image get's loaded depends on the implementation
        of the used ImageLoader class.

        Args:
            path (str): the path to the image which should be loaded
            upper_bound_size (tuple of (width: int, height: int)):
                the maximal size to load data for
            post_load_processor (PostLoadImageProcessor):
                allows to apply changes to the recently loaded image

        Returns:
            ImageHolder: which the image will be assigned to
        """
        raise NotImplementedError()

    def register_error_handler(self, error_handler):
        """Set's the error handler to the passed function.
        An error handler will be called with exceptions which were
        raised during loading an image.

        Args:
            error_handler (Function(Exception)):
                the function which should be called
                to handle an error
        """
        self.error_handler = error_handler

    def process_error(self, exception):
        """Processes an exception.
        Calls the error_handler with the exception
        if there's any.

        Args:
            exception (Exception): the occurred error
        """
        if self.error_handler is not None and exception is not None:
            self.error_handler(exception)

    def __enter__(self):
        pass

    def __exit__(self, *_):
        """Finalises the image loader."""
        pass


class SynchronousImageLoader(ImageLoader):
    """Implementation of ImageLoader
    which loads images right away in the same thread
    it was requested to load the image.
    """

    @staticmethod
    def get_loader_name():
        return "synchronous"

    def load(self, path, upper_bound_size, post_load_processor=None):
        image = None

        try:
            image, _ = load_image(path, None)
        except OSError as exception:
            self.process_error(exception)

        if image and post_load_processor:
            image = post_load_processor.on_loaded(image)

        return ImageHolder(path, image or self.PLACEHOLDER)


class AsynchronousImageLoader(ImageLoader):
    """Extension of ImageLoader
    which adds basic functionality
    needed to implement asynchron image loading.
    """

    @enum.unique
    class Priority(enum.Enum):
        """Enum which defines the possible priorities
        of queue entries.
        """

        HIGH = enum.auto()
        LOW = enum.auto()

    def __init__(self):
        super().__init__()
        self.__queue = queue.Queue()
        self.__queue_low_priority = queue.Queue()
        self.__waiter_low_priority = threading.Condition()

    def _enqueue(
        self, queue, image_holder, upper_bound_size, post_load_processor
    ):
        """Enqueues the image holder weakly referenced.

        Args:
            queue (queue.Queue): the queue to operate on
            image_holder (ImageHolder):
                the image holder for which an image should be loaded
            upper_bound_size (tuple of (width: int, height: int)):
                the maximal size to load data for
            post_load_processor (PostLoadImageProcessor):
                allows to apply changes to the recently loaded image
        """
        queue.put(
            (weakref.ref(image_holder), upper_bound_size, post_load_processor)
        )

    def _dequeue(self, queue):
        """Removes queue entries till an alive reference was found.
        The referenced image holder will be returned in this case.
        Otherwise if there wasn't found any alive reference
        None will be returned.

        Args:
            queue (queue.Queue): the queue to operate on

        Returns:
            tuple of (ImageHolder, tuple of (width: int, height: int),
                      PostLoadImageProcessor):
                an queued image holder or None, upper bound size or None,
                the post load image processor or None
        """
        holder_reference = None
        image_holder = None
        upper_bound_size = None
        post_load_processor = None

        while not queue.empty():
            holder_reference, upper_bound_size, post_load_processor = (
                queue.get_nowait()
            )
            image_holder = holder_reference and holder_reference()
            if holder_reference is None or image_holder is not None:
                break

        return image_holder, upper_bound_size, post_load_processor

    @abc.abstractmethod
    def _schedule(self, function, priority):
        """Schedules the execution of a function.
        Functions should be executed in different thread pools
        based on their priority otherwise you can wait for a death lock.

        Args:
            function (Function): the function which should be executed
            priority (AsynchronImageLoader.Priority):
                the priority of the execution of this function
        """
        raise NotImplementedError()

    def _load_image(self, path, upper_bound_size, post_load_processor):
        """Wrapper for calling load_image.
        Behaves like calling it directly,
        but allows e.g. executing the function in other processes.
        """
        image, *other_data = load_image(path, upper_bound_size)

        if image and post_load_processor:
            image = post_load_processor.on_loaded(image)

        return (image, *other_data)

    def load(self, path, upper_bound_size, post_load_processor=None):
        holder = ImageHolder(path)
        self._enqueue(
            self.__queue, holder, upper_bound_size, post_load_processor
        )
        self._schedule(self.__process_high_priority_entry, self.Priority.HIGH)
        return holder

    def __wait_for_main_work(self):
        """Waits till all queued high priority entries were processed."""
        if not self.__queue.empty():
            with self.__waiter_low_priority:
                if not self.__queue.empty():
                    self.__waiter_low_priority.wait()

    def __notify_main_work_done(self):
        """Notifies waiting threads that
        all queued high priority entries were processed.
        """
        if self.__queue.empty():
            with self.__waiter_low_priority:
                if self.__queue.empty():
                    self.__waiter_low_priority.notify_all()

    def __process_high_priority_entry(self):
        """Processes a single queued high priority entry."""
        self.__process_queue(self.__queue)
        self.__notify_main_work_done()

    def __process_low_priority_entry(self):
        """Processes a single queued low priority entry."""
        self.__wait_for_main_work()
        self.__process_queue(self.__queue_low_priority)

    def __process_queue(self, queue):
        """Processes a single queued entry.

        Args:
            queue (queue.Queue): the queue to operate on
        """
        image = None
        image_holder, upper_bound_size, post_load_processor = self._dequeue(
            queue
        )
        if image_holder is None:
            return

        try:
            image, downscaled = self._load_image(
                image_holder.path, upper_bound_size, post_load_processor
            )
            if upper_bound_size and downscaled:
                self._enqueue(
                    self.__queue_low_priority,
                    image_holder,
                    None,
                    post_load_processor,
                )
                self._schedule(
                    self.__process_low_priority_entry, self.Priority.LOW
                )
        except OSError as exception:
            self.process_error(exception)
        finally:
            image_holder.reveal_image(image or self.PLACEHOLDER)


# * Pythons GIL limits the usefulness of threads.
# So in order to use all cpu cores (assumed GIL isn't released)
# you need to use multiple processes.
# * Pillows load method will read & decode the image.
# So it does the I/O and CPU work.
# Decoding seems to be the bottleneck for large images.
# * Using multiple processes comes with it's own bottleneck
# (transfering the data between the processes).
#
# => Using multiple processes seems to be faster for small images.
#    Using threads seems to be faster for large images.
class ThreadImageLoader(AsynchronousImageLoader):
    """Implementation of AsynchronImageLoader
    which loads images in multiple threads.
    """

    @staticmethod
    def get_loader_name():
        return "thread"

    def __init__(self):
        super().__init__()
        threads = os.cpu_count()
        threads_low_priority = max(1, threads // 2)
        self.__executor = thread.DaemonThreadPoolExecutor(max_workers=threads)
        self.__executor_low_priority = thread.DaemonThreadPoolExecutor(
            max_workers=threads_low_priority
        )
        self.threads = threads + threads_low_priority

    def __exit__(self, *_):
        self.__executor_low_priority.shutdown()
        self.__executor.shutdown()

    def _schedule(self, function, priority):
        executor = self.__executor
        if priority == self.Priority.LOW:
            executor = self.__executor_low_priority
        executor.submit(function).add_done_callback(
            lambda future: self.process_error(future.exception())
        )


class ProcessImageLoader(ThreadImageLoader):
    """Implementation of AsynchronImageLoader
    which loads images in multiple processes.
    Therefore it allows to utilise all cpu cores
    for decoding an image.
    """

    @staticmethod
    def get_loader_name():
        return "process"

    def __init__(self):
        super().__init__()
        self.__executor_loader = concurrent.futures.ProcessPoolExecutor(
            max_workers=self.threads
        )
        # ProcessPoolExecutor won't work
        # when used first in ThreadPoolExecutor
        self.__executor_loader.submit(id, id).result()

    def __exit__(self, *args):
        super().__exit__(*args)
        self.__executor_loader.shutdown()

    @staticmethod
    def _load_image_extern(path, upper_bound_size, post_load_processor):
        """This function is a wrapper for the image loading function
        as sometimes pillow restores decoded images
        received from other processes wrongly.
        E.g. a PNG is reported as webp (-> crash on using an image function)
        So this function is a workaround which prevents these crashs to happen.
        """
        image, *other_data = load_image(path, upper_bound_size)

        if image and post_load_processor:
            image = post_load_processor.on_loaded(image)

        return (image.mode, image.size, image.tobytes(), *other_data)

    def _load_image(self, path, upper_bound_size, post_load_processor=None):
        import PIL.Image

        future = self.__executor_loader.submit(
            ProcessImageLoader._load_image_extern,
            path,
            upper_bound_size,
            post_load_processor,
        )
        mode, size, data, downscaled = future.result()
        return PIL.Image.frombytes(mode, size, data), downscaled


@enum.unique
class ImageLoaderOption(str, enum.Enum):
    """Enum which lists the useable ImageLoader classes."""

    SYNCHRONOUS = SynchronousImageLoader
    THREAD = ThreadImageLoader
    PROCESS = ProcessImageLoader

    def __new__(cls, loader_class):
        inst = str.__new__(cls)
        # Based on an official example
        # https://docs.python.org/3/library/enum.html#using-a-custom-new
        # So.. stfu pylint
        # pylint: disable=protected-access
        inst._value_ = loader_class.get_loader_name()
        inst.loader_class = loader_class
        return inst
